/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.List;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.FutureTask;

/**
 * RunCommand is used to execute a system command and direct the output. It
 * contains only the static run method, which takes a ProcessBuilder object and
 * runs it, possibly redirected the output and error streams to a specified
 * OutputStream object, or returning output in a string. Any stream can also
 * have each line prepended with a tag to indicate the external program being
 * run.
 */
public class RunCommand {

  /**
   * Runs the specified process and returns a string containing the output and
   * error stream results.
   * 
   * @param pb
   *          The ProcessBuilder object containing the command to run
   * @return The output and error stream results of the process
   * @throws IOException
   *           If there is a problem redirecting the streams
   * @throws InterruptedException
   *           If there is a problem redirecting the streams
   * @throws ExecutionException
   *           If there is a problem redirecting the streams
   */
  public static String run(ProcessBuilder pb) throws IOException, InterruptedException, ExecutionException {
    pb.redirectErrorStream(true);
    ByteArrayOutputStream result = new ByteArrayOutputStream();
    Process p = pb.start();
    FutureTask<Boolean> redirector = new FutureTask<Boolean>(new StreamRedirector(p.getInputStream(), result));
    new Thread(redirector).start();
    redirector.get();
    p.waitFor();
    return result.toString();
  }

  /**
   * Runs the specified process and redirects both the output and error streams
   * to the specified OutputStream object.
   * 
   * @param pb
   *          The ProcessBuilder object containing the command to run
   * @param combined
   *          The OutputStream to redirect both the output and error streams to
   * @throws IOException
   *           If there is a problem redirecting the streams
   * @throws InterruptedException
   *           If there is a problem redirecting the streams
   * @throws ExecutionException
   *           If there is a problem redirecting the streams
   */
  public static void run(ProcessBuilder pb, OutputStream combined) throws IOException, InterruptedException,
      ExecutionException {
    run(pb, combined, null, null, null);
  }

  /**
   * Runs the specified process and redirects both the output and error streams
   * to the specified OutputStream object.
   * 
   * @param pb
   *          The ProcessBuilder object containing the command to run
   * @param combined
   *          The OutputStream to redirect both the output and error streams to
   * @param combinedTag
   *          The tag string to prepend to the redirected content
   * @throws IOException
   *           If there is a problem redirecting the streams
   * @throws InterruptedException
   *           If there is a problem redirecting the streams
   * @throws ExecutionException
   *           If there is a problem redirecting the streams
   */
  public static void run(ProcessBuilder pb, OutputStream combined, String combinedTag) throws IOException,
      InterruptedException, ExecutionException {
    run(pb, combined, combinedTag, null, null);
  }

  /**
   * Runs the specified process and redirects both the output and error streams
   * to their respective specified OutputStream objects.
   * 
   * @param pb
   *          The ProcessBuilder object containing the command to run
   * @param stout
   *          The OutputStream to redirect the output stream to
   * @param sterr
   *          The OutputStream to redirect the error stream to
   * @throws IOException
   *           If there is a problem redirecting the streams
   * @throws InterruptedException
   *           If there is a problem redirecting the streams
   * @throws ExecutionException
   *           If there is a problem redirecting the streams
   */
  public static void run(ProcessBuilder pb, OutputStream stdout, OutputStream stderr) throws IOException,
      InterruptedException, ExecutionException {
    run(pb, stdout, null, stderr, null);
  }

  /**
   * Runs the specified process and redirects both the output and error streams
   * to their respective specified OutputStream objects.
   * 
   * @param pb
   *          The ProcessBuilder object containing the command to run
   * @param stout
   *          The OutputStream to redirect the output stream to
   * @param outTag
   *          The tag string to prepend to the redirected output stream content
   * @param sterr
   *          The OutputStream to redirect the error stream to
   * @throws IOException
   *           If there is a problem redirecting the streams
   * @throws InterruptedException
   *           If there is a problem redirecting the streams
   * @throws ExecutionException
   *           If there is a problem redirecting the streams
   */
  public static void run(ProcessBuilder pb, OutputStream stdout, String outTag, OutputStream stderr)
      throws IOException, InterruptedException, ExecutionException {
    run(pb, stdout, outTag, stderr, null);
  }

  /**
   * Runs the specified process and redirects both the output and error streams
   * to their respective specified OutputStream objects.
   * 
   * @param pb
   *          The ProcessBuilder object containing the command to run
   * @param stout
   *          The OutputStream to redirect the output stream to
   * @param sterr
   *          The OutputStream to redirect the error stream to
   * @param errTag
   *          The tag string to prepend to the redirected error stream content
   * @throws IOException
   *           If there is a problem redirecting the streams
   * @throws InterruptedException
   *           If there is a problem redirecting the streams
   * @throws ExecutionException
   *           If there is a problem redirecting the streams
   */
  public static void run(ProcessBuilder pb, OutputStream stdout, OutputStream stderr, String errTag)
      throws IOException, InterruptedException, ExecutionException {
    run(pb, stdout, null, stderr, errTag);
  }

  /**
   * Runs the specified process and redirects both the output and error streams
   * to their respective specified OutputStream objects.
   * 
   * @param pb
   *          The ProcessBuilder object containing the command to run
   * @param stout
   *          The OutputStream to redirect the output stream to
   * @param outTag
   *          The tag string to prepend to the redirected output stream content
   * @param sterr
   *          The OutputStream to redirect the error stream to
   * @param errTag
   *          The tag string to prepend to the redirected error stream content
   * @throws IOException
   *           If there is a problem redirecting the streams
   * @throws InterruptedException
   *           If there is a problem redirecting the streams
   * @throws ExecutionException
   *           If there is a problem redirecting the streams
   */
  public static void run(ProcessBuilder pb, OutputStream stdout, String outTag, OutputStream stderr, String errTag)
      throws IOException, InterruptedException, ExecutionException {

    // Redirect error stream to output stream if no error output specified
    if (stderr == null) {
      pb.redirectErrorStream(true);
    }

    // Print out the command being executed
    System.out.println("[SHEAR] -------------------------");
    List<String> tokens = pb.command();
    System.out.print("[SHEAR] COMMAND: ");
    for (int i = 0; i < tokens.size(); i++) {
      System.out.print(tokens.get(i) + " ");
    }
    System.out.println();
    System.out.println("[SHEAR] -------------------------");

    // Start the process
    Process p = pb.start();

    // Redirect the output stream
    FutureTask<Boolean> redirector = new FutureTask<Boolean>(new StreamRedirector(p.getInputStream(), stdout, outTag));
    new Thread(redirector).start();

    // Redirect the error stream, and wait for it
    if (stderr != null) {
      FutureTask<Boolean> errRedirector = new FutureTask<Boolean>(new StreamRedirector(p.getErrorStream(), stderr,
          errTag));
      new Thread(errRedirector).start();
      errRedirector.get();
    }

    // Wait for the output stream and for the process to finish
    redirector.get();
    p.waitFor();

    System.out.println("[SHEAR] -------------------------");

  }

}
